/*
 * tuc.c - Text to Unix Conversion
 *
 * Copyright 1998 G. Adam Stanislav
 * All rights reserved
 *
 * Permission is granted to distribute this
 * program with FreeBSD providing it is
 * distributed with this source code
 * and the above copyright notice intact.
 *
 * For other distributions, contact me at
 * adam@whizkidtech.net
 *
 * WARNING: I forward ALL SPAM to the abuse
 *          department of spammer's ISP.
 *          Don't even try!
 *
 * This program converts a text file created by any OS
 * into a Unix-style text file in which lines end with
 * a line feed.
 *
 * The simplest way to do this would be to assume
 * that a text file either ends a line with cr/lf
 * or just lf. Then we could just strip any cr's
 * and be done.
 *
 * Alas, that is not a safe assumption. Several
 * methods of line endings exist:
 *
 *    cr/lf
 *    lf
 *    lf/cr
 *    cr
 *
 * Some files even use a combination of two methods
 * (e.g. cr/lf, lf, for two lines in a row).
 *
 * However, searching for the four methods described
 * and convert any of them to lf will work with
 * combination files as well.
 *
 * This program is designed as a state machine.
 * Three states are possible: PROCESSCHAR, IGNORELF, and
 * IGNORECR. Current state depends on the previous
 * state and on the last character processed.
 *
 * The initial state is PROCESSCHAR.
 *
 * This program would best be written in Verilog,
 * but I am not aware of any Verilog compiler,
 * and making a dedicated hardware circuit would
 * be an overkill.
 *
 * The state machine:
 * ===================================================
 * current state | current char | action | next state
 * ===================================================
 *  PROCESSCHAR  | cr           | lf out | IGNORELF
 *               | lf           | lf out | IGNORECR
 *               | c (other)    |  c out | PROCESSCHAR
 * ---------------------------------------------------
 *  IGNORELF     | cr           | lf out | IGNORELF
 *               | lf           | none   | PROCESSCHAR
 *               | c (other)    | c out  | PROCESSCHAR
 * ---------------------------------------------------
 *  IGNORECR     | cr           | none   | PROCESSCHAR
 *               | lf           | lf out | IGNORECR
 *               | c (other)    | c out  | PROCESSCHAR
 * ===================================================
 *
 * We could use this table to write the code directly.
 * It is, however, worth noticing that the action and
 * the next state for the default (marked "c (other)"
 * under "current char") is always the same, regardless
 * of current state.
 *
 * Hence, it makes for more efficient code to examine
 * the second column first.
 *
 * usage: tuc [-d|D ][[-i] inputfile][[-o] outputfile]
 *
 * Version 1.00 - 26 December 1998
 * Version 1.10 - same day
 *
 * I do not know why it did not occur to me when doing
 * version 1.00 - new in 1.10 is the -d or -D switch which
 * reverses the effect of tuc, i.e., all lines in
 * the output end with cr/lf. This is useful when
 * converting Unix text file for use under DOS.
 *
 * The difference between the two: -d is ignored if
 * output goes to stdout, -D is not ignored. Note
 * that you can select one or the other but not both.
 *
 * Why would you want -d ignored when writing to stdout?
 * Because under MS DOS stdout is opened in "cooked"
 * mode. That means that MS DOS converts the lf into cr/lf
 * when writing to stdout. Using the -D option will result
 * in ending each line with cr/cr/lf, but only when writing
 * to stdout!
 *
 * So use -d when running this program under MS DOS, -D when
 * running it under Unix.
 *
 */
#include <stdio.h>

static int usage(FILE *input, FILE *output);
static int cantopen(char *filename, FILE *input, FILE *output);

typedef enum {
	PROCESSCHAR,
	IGNORELF,
	IGNORECR
} TUCSTATE;

int main(int argc, char *argv[]) {
	int i;
	int gotinput;
	int gotoutput;
	int dos;
	int register c;
	char *strptr;
	FILE *input;
	FILE *output;
	TUCSTATE register state;

	gotinput = gotoutput = dos = 0;
	input    = output    = NULL;

	for (i = 1; i < argc; i++) {
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
			case 'i':
				if (argv[i][2] != '\0')
					strptr = &(argv[i][2]);
				else if (++i < argc)
					strptr = argv[i];
				else return usage(input, output);
				if (gotinput++)
					return usage(input, output);
				input = fopen(strptr, "rb");
				if (input == NULL)
					return cantopen(strptr, NULL, output);
				break;
			case 'o':
				if (argv[i][2] != '\0')
					strptr = &(argv[i][2]);
				else if (++i < argc)
					strptr = argv[i];
				else return usage(input, output);
				if (gotoutput++)
					return usage(input, output);
				output = fopen(strptr, "wb");
				if (output == NULL)
					return cantopen(strptr, input, NULL);
				break;
			case 'd':
				if ((argv[i][2] != '\0') || (dos++ != 0))
					return usage(input, output);
				break;
			case 'D':
				if ((argv[i][2] != '\0') || (dos-- != 0))
					return usage(input, output);
				break;
			default:
				return usage(input, output);
			}
		}
		else if (gotinput++ == 0) {
			input = fopen(argv[i], "rb");
			if (input == NULL)
				return cantopen(argv[i], NULL, output);
		}
		else if (gotoutput++ == 0) {
			output = fopen(argv[i], "wb");
			if (output == NULL)
				return cantopen(argv[i], input, NULL);
		}
		else return usage(input, output);
	}

	if (gotinput == 0)
		input = stdin;

	if (gotoutput == 0) {
		output = stdout;
		if (dos > 0)
			dos = 0;
	}

	state = PROCESSCHAR;

	while (!feof(input)) {
		c = fgetc(input);
		switch (c) {
		case '\n' :
			if (state == IGNORELF)
				state = PROCESSCHAR;
			else {
				if (dos)
					fputc('\r', output);
				fputc('\n', output);
				state = IGNORECR;
			}
			/* fall through */
		case EOF:
			break;
		case '\r':
			if (state == IGNORECR)
				state = PROCESSCHAR;
			else {
				if (dos)
					fputc('\r', output);
				fputc('\n', output);
				state = IGNORELF;
			}
			break;
		default:
			fputc(c, output);
			state = PROCESSCHAR;
			break;
		}
	}

	if (gotinput != 0)
		fclose(input);
	if (gotoutput != 0)
		fclose(output);

	return 0;
}

static int usage(FILE *input, FILE *output) {

	if (input != NULL)
		fclose(input);

	if (output != NULL)
		fclose(output);

	fprintf(stderr, "Usage: tuc [-d|D ][[-i] inputfile][[ -o][ outputfile]\n");
	return 1;
}

static int cantopen(char *filename, FILE *input, FILE *output) {

	if (input != NULL)
		fclose(input);

	if (output != NULL)
		fclose(output);

	fprintf(stderr, "Can't open %s.\n", filename);
	return 2;
}
